
/*
 * Performance event assistance - Pointer chasing test
 */

/* 
 * Author:  P.J. Drongowski
 * Site:    sandsoftwaresound.net
 * Version: 1.0
 * Date:    27 January 2021
 *
 * Copyright (c) 2021 Paul J. Drongowski
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include "pe_assist.h"
#include "pe_cortex_a72.h"

/*
 * Test parameters
 */
#define TEST_ITERATIONS       (4096 * 1024)

int iteration_count = TEST_ITERATIONS ;

/*
 * Default values for the pointer array and the cache
 * line size. See chase.c for descriptions of these
 * data structures.
 */
#define ARRAY_ELEMENTS     (2048)
#define POINTER_SIZE       (sizeof(void*))
#define LINE_SIZE          64
#define POINTERS_PER_LINE  (LINE_SIZE / POINTER_SIZE)

typedef struct _CacheLine {
  struct _CacheLine* ptrCacheLine[POINTERS_PER_LINE] ;
} CacheLine ;

long int array_elements = ARRAY_ELEMENTS ;
long int line_size = LINE_SIZE ;
long int pointer_size = POINTER_SIZE ;
long int pointers_per_line = POINTERS_PER_LINE ;

/*
 * The program allocates an array of CacheLine structures. The number
 * of elements (cache lines) is determined by the variable array_elements.
 * The array is initialized such that it forms a linked list of elements.
 * Elements in the upper half of the array point to elements in the
 * lower half, and vice versa. Thus, access will ping-pong between
 * the upper and lower halves when the linked list is walked (pointers
 * are chased.)
 */
CacheLine* array = NULL ;

/*
 * Allocate and initialize the linked list array
 */
void initialize_array()
{
  long int steps ;   // Number of initialization setps
  long int top ;     // Index of top cache line
  long int bottom ;  // Index of bottom cache line
  long int i ;       // Array index counter
  int j ;            // Pointer index counter

  if ((array = (CacheLine*) malloc(array_elements * line_size)) == NULL) {
    fprintf(stderr, "Couldn't allocate test array\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Set the entire array to zero
  for (i = 0 ; i < array_elements ; i++) {
    for (j = 0 ; j < pointers_per_line ; j++) {
      array[i].ptrCacheLine[j] = NULL ;
    }
  }

  // The test array is divided into two halves: top and
  // bottom. Set the pointers
  // so that memory accesses ping-pong between the two
  // halves of the array. Ping-pong between cache lines.
  steps = array_elements / 2 ;
  bottom = 0 ;
  top = steps ;
  for (i = 0 ; i < steps ; i++) {
    array[bottom].ptrCacheLine[0] = &array[top] ;
    array[top].ptrCacheLine[0] = &array[bottom+1] ;
    top++ ;
    bottom++ ;
  }

  // Terminate the linked list
  array[(array_elements-1)].ptrCacheLine[0] = NULL ;
}

/*
 * Pointer chasing loop. This is the workload without instrumentation.
 */
void pointer_chase_loop(CacheLine* array)
{
  CacheLine* p ;
  for(p = array ; p != NULL ; p = p->ptrCacheLine[0]) ;
}

/*
 * Instrumented pointer chasing loop. This is the workload.
 */
void instrumented_chase_loop(CacheLine* array)
{
  CacheLine* p ;
  for(p = array ; p != NULL ; p = p->ptrCacheLine[0]) ;
}

int main(int argc, char* argv[])
{
  double finish_time = 0.0 ;
  
  // Allocate and initialize the linked list
  initialize_array() ;

  // Measure events related to data read access
  a72MeasureDataAccessEvents() ;

  // Clear and start the counters and clock
  start_clock() ;
  peStartCounting() ;

  // Perform pointer chasing
  for (register int i = iteration_count ; i > 0 ; i--) {
    // pointer_chase_loop(array) ;
    for(CacheLine* p = array ; p != NULL ; p = p->ptrCacheLine[0]) ;
  }

  // Stop the counters
  peStopCounting() ;
  finish_time = get_clock_time() ;
  
  // Display run parameters
  fprintf(stdout,"Memory access test\n") ;
  fprintf(stdout,"  Number of test iterations:  %16d\n", iteration_count) ;
  fprintf(stdout,"  List items/cache lines:     %16d\n", array_elements) ;
  fprintf(stdout,"  Array size/bytes:           %16d\n", array_elements * line_size) ;
  fprintf(stdout,"  Cache line size:            %16d\n", line_size) ;
  fprintf(stdout,"  Pointer size:               %16d\n", pointer_size) ;
  fprintf(stdout,"  Pointers per line:          %16d\n", pointers_per_line) ;

  // Display performance measurements
  fprintf(stdout,"Performance measurements\n") ;
  print_clock_time(stdout, finish_time) ;
  a72PrintDataAccessEvents(stdout) ;
}
