
/*
 * Performance event assistance - Branch mispredicts test
 */

/* 
 * Author:  P.J. Drongowski
 * Site:    sandsoftwaresound.net
 * Version: 1.1
 * Date:    14 January 2021
 *
 * Copyright (c) 2021 Paul J. Drongowski
 */

#include "pe_assist.h"
#include "pe_cortex_a72.h"

/*
 * Test parameters
 */
#define TEST_ITERATIONS       200 
#define ARRAY_SIZE            20000000 

/*
 * Walk through the array one character at a time. If a
 * character is equal to '1', add 1 to the running count.
 * Return the final count.
 */
int test_loop(int size, char* array)
{
    register int count = 0 ;
    register char* r = array ;
    int s ;

    for (s = size ; s > 0 ; s--) 
    {
        if (*r++ == '1') count++ ;
    }
    return( count ) ;
}

/*
 * Initialize the test array such that 50% of the
 * characters will be '0' and 50% will be '1'. The
 * always_one and always_zero flags force the content
 * of the test array to '1' or '0', respectively.
 */
void initialize_test_array(int size, char* array,
			   int always_one, int always_zero)
{
    register char* r = array ;
    int s ;
    for (s = size ; s > 0 ; s--) 
    {
	if (always_one)
	{
            *r++ = '1' ;
	} else if (always_zero)
	{
            *r++ = '0' ;
	} else {
            *r++ = ((rand() & 0x1) ? '1' : '0') ;
	}
    }
}

/*
 * Perform the test itself. Since all of the test-related
 * activity starts in this function, it should be easier to
 * track samples back to the test itself.
 */
void run_the_test(int iteration_count, int array_size, char* array)
{
  register int rarray_size = array_size ;
  register char* rarray = array ;
  int i ;

  for (i = iteration_count ; i-- ; ) {
      test_loop(array_size, array) ;
  }
}

/*
 * Test driver
 */
int main(int argc, char **argv)
{
  int iteration_count = TEST_ITERATIONS ;
  char* array = NULL ;
  int array_size = ARRAY_SIZE ;
  int always_one = 0 ;
  int always_zero = 0 ;

  // Allocate and initialize the test array
  array = (char*)malloc((size_t)array_size) ;
  initialize_test_array(array_size, array, always_one, always_zero) ;
  
  fprintf(stdout, "*************************************\n") ;
  fprintf(stdout, "Mispredicts: Instruction events\n") ;
  fprintf(stdout, "Cortex-A72 native events\n");
  if (always_one) {
    fprintf(stdout, "Always one\n") ;
  } else if (always_zero) {
    fprintf(stdout, "Always zero\n") ;
  } else {
    fprintf(stdout, "Randomized one and zero\n") ;
  }
  fprintf(stdout, "*************************************\n") ;

  // Measure events related to instructions and branches
  start_clock() ;
  a72MeasureInstructionEvents() ;
  
  // Clear and start the counters
  peStartCounting() ;

  // Do the workload
  run_the_test(iteration_count, array_size, array) ;

  // Stop the counters
  peStopCounting() ;

  // Print results
  print_clock_time(stdout, get_clock_time()) ;
  a72PrintInstructionEvents(stdout) ;

  return( 0 ) ;
}
