
/*
 * Performance event assistance functions
 */

/* 
 * Author:  P.J. Drongowski
 * Site:    sandsoftwaresound.net
 * Version: 1.0
 * Date:    19 November 2020
 *
 * Copyright (c) 2020 Paul J. Drongowski
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/perf_event.h>
#include <asm/unistd.h>

#include "pe_assist.h"

static int           peNumberOfGroups = 0 ;     // For now, one group maximum
static int           peNumberOfEvents = 0 ;     // Number of events in the group
static int           peLeaderDefined  = 0 ;     // TRUE if leader is defined
static pid_t         pePID = 0 ;                // The process to be monitored
static int           peCPU = -1 ;               // The CPU to be monitored
static unsigned long peFlags = 0 ;              // Event monitoring flags
static int           peGroupFD = 0 ;            // Event group file descriptor
static int           peEventFD[PE_MAX_EVENTS] ; // Event file descriptors

static struct perf_event_attr peAttr ;          // Working event attributes
static int           peEventIndex = 0 ;         // Working event array index

/*
 * Linux performance event system call
 */
static long
perf_event_open(struct perf_event_attr *hw_event, pid_t pid,
                int cpu, int group_fd, unsigned long flags)
{
    int ret;

    ret = syscall(__NR_perf_event_open, hw_event, pid, cpu, group_fd, flags);
    return ret;
}

/*
 * Initialize internal data structures. Must be called first.
 */
int peInitialize()
{
  int i ;
  
  peNumberOfGroups = 0 ;
  peNumberOfEvents = 0 ;
  peLeaderDefined = 0 ;
  pePID = 0 ;
  peCPU = -1 ;
  peFlags = 0 ;
  peGroupFD = 0 ;
  for (i = 0 ; i < PE_MAX_EVENTS ; i++) peEventFD[i] = 0 ;

  peEventIndex = 0 ;
  
  return( 1 ) ;
}

/*
 * Make a performance event/monitoring group. Define the number of events in
 * the group including the leader. All events in the group have the same
 * PID, CPU and flag parameter values.
 *    num: Number of events in the group (including the leader)
 *    pid: Process ID parameter
 *    cpu: CPU parameter (-1 is all)
 *    flags: perf_event_open() flags parameter
 */
int peMakeGroup(int num, pid_t pid, int cpu, unsigned long flags)
{
  if (peNumberOfGroups >= PE_MAX_GROUPS) {
    fprintf(stderr, "Too many groups\n") ;
    return( 0 ) ;
  }
  if (num > PE_MAX_EVENTS) {
    fprintf(stderr, "Too many events requested for group") ;
    return( 0 ) ;
  }
  peNumberOfGroups++ ;
  peNumberOfEvents = num ;
  peEventIndex = 0 ;
  peLeaderDefined = 0 ;
  pePID = pid ;
  peCPU = cpu ;
  peFlags = flags ;
  return( 1 ) ;
}

/*
 * Add event leader to an existing event group.
 *    etype: Perf event type
 *    econfig: Perf event identifier
 */
int peAddLeader(long unsigned etype, long unsigned econfig)
{
  if (peNumberOfGroups == 0) {
    fprintf(stderr, "No group defined (peAddLeader)\n") ;
    return( 0 ) ;
  }
  if (peLeaderDefined) {
    fprintf(stderr, "Group event leader already defined") ;
    return( 0 ) ;
  }
  if (peEventIndex >= peNumberOfEvents) {
    fprintf(stderr, "Too many events (peAddLeader)\n") ;
    return( 0 ) ;
  }

  memset(&peAttr, 0, sizeof(struct perf_event_attr)) ;
  peAttr.type = etype ;
  peAttr.size = sizeof(struct perf_event_attr) ;
  peAttr.config = econfig ;
  peAttr.disabled = 1 ;
  peAttr.exclude_kernel = 1 ;
  peAttr.exclude_hv = 1 ;

  peGroupFD = perf_event_open(&peAttr, pePID, peCPU, -1, peFlags) ;
  if (peGroupFD == -1) {
    perror("peAddLeader") ;
    fprintf(stderr, "Error opening leader %llx\n", peAttr.config) ;
    exit(EXIT_FAILURE) ;
  }

  // Save the event file descriptor and advance the index
  peEventFD[peEventIndex] = peGroupFD ;
  peEventIndex++ ;
  peLeaderDefined = 1 ;
  return( 1 ) ;
}

/*
 * Add event to a group with an existing leader
 */
int peAddEvent(long unsigned etype, long unsigned econfig)
{
  int efd = 0 ;

  if (peNumberOfGroups == 0) {
    fprintf(stderr, "No group defined (peAddEvent)\n") ;
    return( 0 ) ;
  }
  if (peLeaderDefined == 0) {
    fprintf(stderr, "Event leader not defined for event group") ;
    return( 0 ) ;
  }
  if (peEventIndex >= peNumberOfEvents) {
    fprintf(stderr, "Too many events (peAddEvent)\n") ;
    return( 0 ) ;
  }

  memset(&peAttr, 0, sizeof(struct perf_event_attr)) ;
  peAttr.type = etype ;
  peAttr.size = sizeof(struct perf_event_attr) ;
  peAttr.config = econfig ;
  peAttr.disabled = 0 ;
  peAttr.exclude_kernel = 1 ;
  peAttr.exclude_hv = 1 ;

  efd = perf_event_open(&peAttr, pePID, peCPU, peGroupFD, peFlags) ;
  if (efd == -1) {
    perror("peAddEvent") ;
    fprintf(stderr, "Error opening event %llx\n", peAttr.config) ;
    exit( EXIT_FAILURE ) ;
  }

  // Save the event file descriptor and advance the index
  peEventFD[peEventIndex] = efd ;
  peEventIndex++ ;
  return( 1 ) ;
}

int peStartCounting()
{
  if (ioctl(peGroupFD, PERF_EVENT_IOC_RESET, 0) < 0) {
    perror("peStartCounting") ;
    fprintf(stderr, "Error when resetting event counters\n") ;
    return( 0 ) ;
  }
  if (ioctl(peGroupFD, PERF_EVENT_IOC_ENABLE, 0) < 0) {
    perror("peStartCounting") ;
    fprintf(stderr, "Error when starting event counters\n") ;
    return( 0 ) ;
  }
  return( 1 ) ;
}

int peStopCounting()
{
  if (ioctl(peGroupFD, PERF_EVENT_IOC_DISABLE, 0) < 0) {
    perror("peStopCounting") ;
    fprintf(stderr, "Error when stopping event counters\n") ;
    return( 0 ) ;
  }
  return( 1 ) ;
}

int peResetCounters()
{
  if (ioctl(peGroupFD, PERF_EVENT_IOC_RESET, 0) < 0) {
    perror("peResetCounters") ;
    fprintf(stderr, "Error when resetting event counters\n") ;
    return( 0 ) ;
  }
  return( 1 ) ;
}

long long peReadCount(int eindex)
{
  long long count ;
  int fd ;

  if ((eindex < 0) || (eindex >= peNumberOfEvents)) {
    fprintf(stderr, "Invalid event index %d (pePrintCount)\n", eindex) ;
    return( 0 ) ;
  }

  fd = peEventFD[eindex] ;
  if (read(fd, &count, sizeof(long long)) < 0) {
    perror("peReadCount") ;
    fprintf(stderr, "Error when reading event count %d\n", eindex) ;
    return( 0 ) ;
  }
  return( count ) ;
}

void pePrintCount(FILE* stream, int eindex, char *legend)
{
  long long count ;

  count = peReadCount(eindex) ;
  fprintf(stream, "%s%lld\n", legend, count) ;
}

int peMeasureInstructionEvents()
{
  // Initialize the event helper module
  peInitialize() ;

  // Make an event group (two events, monitor this PID, monitor all CPUs)
  // Request 6 events, monitor this PID, monitor all CPUs
  if (peMakeGroup(6, 0, -1, 0) == 0) {
    fprintf(stderr, "Couldn't make event group (main)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add leader event to the group
  if (peAddLeader(PERF_TYPE_HARDWARE, PERF_COUNT_HW_INSTRUCTIONS) == 0) {
    fprintf(stderr, "Couldn't add event group leader (instructions)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_CPU_CYCLES) == 0) {
    fprintf(stderr, "Couldn't add event to the group (cpu cycles)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_BRANCH_INSTRUCTIONS) == 0) {
    fprintf(stderr, "Couldn't add event to the group (branches)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_BRANCH_MISSES) == 0) {
    fprintf(stderr, "Couldn't add event to the group (branch misses)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_CACHE_REFERENCES) == 0) {
    fprintf(stderr, "Couldn't add event to the group (cache refs)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_CACHE_MISSES) == 0) {
    fprintf(stderr, "Couldn't add event to the group (cache misses)\n") ;
    exit( EXIT_FAILURE ) ;
  }
  return( 1 ) ;
}

int pePrintInstructionEvents(FILE* stream)
{
  // Read instruction event counts
  long long instructions  = peReadCount(0) ;
  long long cpu_cycles    = peReadCount(1) ;
  long long branches      = peReadCount(2) ;
  long long branch_misses = peReadCount(3) ;
  long long cache_refs    = peReadCount(4) ;
  long long cache_misses  = peReadCount(5) ;

  // Print event counts
  fprintf(stream, "Instructions:  %lld\n", instructions) ;
  fprintf(stream, "CPU cycles:    %lld\n", cpu_cycles) ;
  fprintf(stream, "Branches:      %lld\n", branches) ;
  fprintf(stream, "Branch misses: %lld\n", branch_misses) ;
  fprintf(stream, "Cache refs:    %lld\n", cache_refs) ;
  fprintf(stream, "Cache misses:  %lld\n", cache_misses) ;
  return( 1 ) ;
}

int peMeasureDataAccessEvents()
{
  // Initialize the event helper module
  peInitialize() ;

  // Make an event group (two events, monitor this PID, monitor all CPUs)
  // Request 6 events, monitor this PID, monitor all CPUs
  if (peMakeGroup(6, 0, -1, 0) == 0) {
    fprintf(stderr, "Couldn't make event group (main)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add leader event to the group
  if (peAddLeader(PERF_TYPE_HARDWARE, PERF_COUNT_HW_INSTRUCTIONS) == 0) {
    fprintf(stderr, "Couldn't add event group leader (instructions)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_CPU_CYCLES) == 0) {
    fprintf(stderr, "Couldn't add event to the group (cpu cycles)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_BRANCH_INSTRUCTIONS) == 0) {
    fprintf(stderr, "Couldn't add event to the group (branches)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_BRANCH_MISSES) == 0) {
    fprintf(stderr, "Couldn't add event to the group (branch misses)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_CACHE_REFERENCES) == 0) {
    fprintf(stderr, "Couldn't add event to the group (cache refs)\n") ;
    exit( EXIT_FAILURE ) ;
  }

  // Add event to the group
  if (peAddEvent(PERF_TYPE_HARDWARE, PERF_COUNT_HW_CACHE_MISSES) == 0) {
    fprintf(stderr, "Couldn't add event to the group (cache misses)\n") ;
    exit( EXIT_FAILURE ) ;
  }
  return( 1 ) ;
}

int pePrintDataAccessEvents(FILE *stream)
{
  // Print event counts
  pePrintCount(stream, 0, "Instructions:  ") ;
  pePrintCount(stream, 1, "CPU cycles:    ") ;
  pePrintCount(stream, 2, "Branches:      ") ;
  pePrintCount(stream, 3, "Branch misses: ") ;
  pePrintCount(stream, 4, "Cache refs:    ") ;
  pePrintCount(stream, 5, "Cache misses:  ") ;
  return( 1 ) ;
}
