
/*
 * Build a cache-line shuffled linked list and chase the pointers
 */

/* 
 * Author:  P.J. Drongowski
 * Site:    sandsoftwaresound.net
 * Version: 1.2
 * Date:    18 February 2021
 *
 * Copyright (c) 2021 Paul J. Drongowski
 */

#include <stdio.h> 
#include <stdlib.h> 
#include <time.h> 

#include "pe_assist.h"
#include "pe_cortex_a72.h"

// 64-byte cache line
typedef struct _CacheLine {
  struct _CacheLine* nextLine ;
  long long int value ;
  long long int padding[6] ;
} CacheLine ;

// Default number of cache line list elements
#define LIST_ELEMENTS   32
long int number_of_elements = LIST_ELEMENTS ;
CacheLine* cacheLineArray = NULL ;
CacheLine* listHead = NULL ;

// Permutation array for shuffing 
long int* permute ;

/*
 * Swap permutation indices
 */
void swap (long int* a, long int* b) 
{
  // Swap the array elements a and b
  long int temp = *a ;
  *a  = *b ;
  *b = temp ;
} 

/*  
 * Print the array in order. For testing purposes.
 */
void printArray(int n, long int array[]) 
{ 
  for (long int i = 0; i < n; i++) {
    printf("%ld ", array[i]) ;
  }
  printf("\n") ;
} 

/*
 * Print the cache line array in order. For testing purposes.
 */
void printLines(int n, CacheLine* array) 
{ 
  for (long int i = 0; i < n; i++) {
    printf("%lld ", array[i].value) ;
  }
  printf("\n") ;
} 

/*
 * Print memory size (K, M or G)
 */
void printMemorySize(FILE* stream, long int size)
{
  if (size >= (1024*1024*1024)) {
    // Giga
    fprintf(stream, "%ldGB\n", (long int)(size/(1024*1024*1024))) ;
  } else if (size >= (1024*1024)) {
    // Mega
    fprintf(stream, "%ldMB\n", (long int)(size/(1024*1024))) ;
  } else if (size >= (1024)) {
    // Kilo
    fprintf(stream, "%ldKB\n", (long int)(size/1024)) ;
  } else {
    fprintf(stream, "%ldB\n", size) ;
  }
}

// Function to print the list by chasing links (testing)
void printList(CacheLine* head)
{
  for (CacheLine* p = head ; p != NULL ; p = p->nextLine)
        printf("%lld ", p->value); 
    printf("\n"); 
}

/*
 * Shuffle the elements in the permutation array.
 */
void shuffle(int n) 
{ 
    // Use the same seed to replicate runs
    srand( 0x373737 ) ;
  
    // Start from the last element and swap one by one. We don't 
    // need to run for the first element that's why i > 0 
    for (int i = n-1; i > 0; i--) 
    { 
        // Pick a random index from 0 to i 
        int j = rand() % (i+1); 
  
        // Swap arr[i] with the element at random index 
        swap(&permute[i], &permute[j]); 
    } 
} 

/*
 * Create the linked list of cache lines. The list elements
 * are shuffled physically within the storage block.
 */
void initialize(int n)
{
  register CacheLine* p = NULL ;
  register long int i, j ;

  // Allocate the permutation array
  permute = (long int*)malloc(n * sizeof(long int)) ;

  // Initialize the permutation array with indices 0 to n-1
  for (long int i = 0 ; i < n ; i++) {
      permute[i] = i ;
  }

  // Shuffle the elements of the permutation array.
  // The shuffled indices determine the physical layout of
  // the cache line elements in storage.
  shuffle(n) ; 
  // DEBUG: printArray(number_of_elements, permute) ; 

  // Allocate storage for the linked list of cache lines
  cacheLineArray = (CacheLine*)malloc(n * sizeof(CacheLine)) ;

  // Create the linked list with values 0 to n-1
  p = cacheLineArray ;
  listHead = &cacheLineArray[permute[0]] ;
  for (i = 0 ; i < (n-1) ; i++) {
    j = permute[i] ;
    p[j].value = i ;
    p[j].nextLine = &cacheLineArray[permute[i+1]] ;
  }
  // Terminate the list
  j = permute[i] ;
  p[j].value = i ;
  p[j].nextLine = NULL ;

  // Free the permutation array
  free(permute) ;
}

/*
 * Test driver
 */
int main(int argc, char* argv[]) 
{
  long int iterations = 1 ;

  if (argc > 1) {
    number_of_elements = atoi(argv[1]) ;
  } else {
    number_of_elements = LIST_ELEMENTS ;
  }
  iterations = (64 * 4194304) / number_of_elements ;
  fprintf(stdout, "Number of list elements: %ld\n", number_of_elements) ;
  fprintf(stdout, "Number of iterations:    %ld\n", iterations) ;

  fprintf(stdout, "Size of CacheLine:  %d\n", sizeof(CacheLine)) ;
  fprintf(stdout, "Storage array size: ") ;
  printMemorySize(stdout, (long int)(number_of_elements*sizeof(CacheLine))) ;
  
  initialize(number_of_elements) ;

#ifdef DATA_CACHE_EVENTS
  a72MeasureDataCacheEvents() ;
#else
  a72MeasureDataAccessEvents() ;
#endif

  // DEBUG: printList(listHead) ;
  // DEBUG: printLines(number_of_elements, cacheLineArray) ;

  // Test loop (workload)
  start_clock() ;
  peStartCounting() ;
  for ( ; iterations > 0 ; iterations--) {
    for (CacheLine *p = listHead ; p != NULL ; p = p->nextLine) ;
  }
  peStopCounting() ;

  print_clock_time(stdout, get_clock_time()) ;
#ifdef DATA_CACHE_EVENTS
  a72PrintDataCacheEvents(stdout) ;
#else
  a72PrintDataAccessEvents(stdout) ;
#endif

  return( 0 ) ; 
} 
